def test_readme_1():
    import cabaret

    ra, dec = 12.3323, 30.4343  # in degrees
    exposure_time = 10  # in seconds

    image = cabaret.generate_image(ra, dec, exposure_time)

    assert image.shape == (
        1024,
        1024,
    ), f"Expected image shape (1024, 1024), but got {image.shape}"


def test_readme_2():
    import cabaret
    from cabaret import Camera

    camera = Camera(read_noise=10, gain=1, width=100, height=100)

    ra, dec = 12.3323, 30.4343  # in degrees
    exposure_time = 10  # in seconds

    image = cabaret.generate_image(ra, dec, exposure_time, camera=camera)

    assert image.shape == (
        100,
        100,
    ), f"Expected image shape (1000, 1000), but got {image.shape}"



import pytest
from astropy.coordinates import SkyCoord

from cabaret.queries import get_gaia_sources


def test_get_gaia_sources_basic():
    center = SkyCoord(ra=10.68458, dec=41.26917, unit="deg")
    fov = 0.05
    sources = get_gaia_sources(center, fov, limit=10, timeout=30)
    assert len(sources) <= 10
    assert sources is not None


def test_get_gaia_sources_timeout():
    center = SkyCoord(ra=10.68458, dec=41.26917, unit="deg")
    fov = 0.05
    with pytest.raises(TimeoutError):
        get_gaia_sources(center, fov, limit=10, timeout=0.0001)



from dataclasses import dataclass


@dataclass
class Site:
    sky_background: float = 150  # for I+z band in Paranal, e-/m2/arcsec2/s
    seeing: float = 1.3  # arcsec
    latitude: float = None  # degrees, optional
    longitude: float = None  # degrees, optional



from cabaret.camera import Camera
from cabaret.focuser import Focuser
from cabaret.image import generate_image
from cabaret.observatory import Observatory
from cabaret.site import Site
from cabaret.sources import Sources
from cabaret.telescope import Telescope

__all__ = [
    "Camera",
    "Focuser",
    "generate_image",
    "Observatory",
    "Site",
    "Sources",
    "Telescope",
]



from dataclasses import dataclass

import numpy as np
from astropy.coordinates import Longitude, SkyCoord
from astropy.wcs import WCS


@dataclass
class Sources:
    """A collection of sources with their sky coordinates and fluxes.

    Parameters
    ----------
    coords : SkyCoords
        A SkyCoords instance of size n, with the ra and dec coordinates of the sources.
    fluxes : np.ndarray
        An array of shape (n,) containing the fluxes of the sources.

    Examples
    --------
    >>> from cabaret.queries import Sources
    >>> import numpy as np
    >>> from astropy.coordinates import SkyCoord
    >>> coords = np.array([[10.64, 41.26], [10.68, 41.22]])
    >>> fluxes = np.array([169435.6, 52203.9])
    >>> sources = Sources(SkyCoord(coords, unit='deg'), fluxes)
    """

    coords: SkyCoord
    fluxes: np.ndarray

    def __post_init__(self):
        if not isinstance(self.coords, SkyCoord):
            raise ValueError("coords must be an instance of SkyCoord.")
        if not isinstance(self.fluxes, np.ndarray):
            try:
                self.fluxes = np.array(self.fluxes)
            except Exception:
                raise ValueError("fluxes must be an instance of np.ndarray.")
        if self.coords.size != self.fluxes.size:
            raise ValueError("coords and fluxes must have the same length.")

    @property
    def ra(self) -> Longitude:
        return self.coords.ra

    @property
    def dec(self) -> Longitude:
        return self.coords.dec

    def to_pixel(self, wcs: WCS) -> np.ndarray:
        """Convert the RA-DEC coordinates to pixel coordinates using the given WCS.

        Parameters
        ----------
        wcs : astropy.wcs.WCS
            The WCS object used for the conversion.

        Returns
        -------
        np.ndarray
            An array of shape (n, 2) containing the pixel coordinates of the sources.
        """
        return np.array(self.coords.to_pixel(wcs))

    def __len__(self) -> int:
        return len(self.fluxes)

    @classmethod
    def from_arrays(
        cls,
        ra: np.ndarray | list,
        dec: np.ndarray | list,
        fluxes: np.ndarray | list,
        units: str = "deg",
    ) -> "Sources":
        """Create a Sources instance from separate RA and DEC arrays.

        Parameters
        ----------
        ra : np.ndarray
            An array of shape (n,) containing the RA coordinates of the sources in deg.
        dec : np.ndarray
            An array of shape (n,) containing the DEC coordinates of the sources in deg.
        fluxes : np.ndarray
            An array of shape (n,) containing the fluxes of the sources.
        **kwargs
            Additional keyword arguments passed to the Sources constructor.

        Returns
        -------
        Sources
            A Sources instance.
        """
        if not isinstance(ra, np.ndarray):
            try:
                ra = np.array(ra)
            except Exception:
                raise ValueError("ra must be an instance of np.ndarray.")
        if not isinstance(dec, np.ndarray):
            try:
                dec = np.array(dec)
            except Exception:
                raise ValueError("dec must be an instance of np.ndarray.")
        if ra.shape != dec.shape:
            raise ValueError("ra and dec must have the same shape.")

        parameters = {
            "coords": SkyCoord(ra=ra, dec=dec, unit=units),
            "fluxes": fluxes,
        }
        return cls(**(parameters))



import numpy as np
import pytest

from cabaret.camera import (
    Camera,
    ConstantPixelDefect,
    RandomNoisePixelDefect,
    TelegraphicPixelDefect,
)


@pytest.fixture
def camera():
    return Camera(height=128, width=128)


@pytest.fixture
def image(camera):
    return np.zeros((camera.height, camera.width), dtype=np.float32)


def test_constant_pixel_defect_initialization():
    defect = ConstantPixelDefect(name="constant_defect", rate=0.1, value=1024)
    assert defect.name == "constant_defect"
    assert defect.rate == 0.1
    assert defect.value == 1024


def test_pixel_defect_number_of_defect_pixels():
    camera = Camera(width=1024, height=1024)
    defect = ConstantPixelDefect(rate=0.1)
    assert defect.number_of_defect_pixels(camera) == np.round(0.1 * 1024 * 1024)


def test_pixel_defect_select_random_pixels(camera):
    defect = ConstantPixelDefect(name="test_defect", rate=0.1, seed=42)
    defect.set_pixels(defect._select_random_pixels(camera), camera)

    expected_number_of_pixels = defect.number_of_defect_pixels(camera)
    assert defect.pixels.shape[0] == expected_number_of_pixels


def test_constant_pixel_defect_introduce_pixel_defect_with_set_pixels(image):
    camera = Camera(width=5, height=5)
    defect = ConstantPixelDefect(value=128, seed=42)

    # Set specific pixels
    specific_pixels = np.array([[1, 1], [2, 2], [3, 3]])
    defect.set_pixels(specific_pixels, camera)

    defect.introduce_pixel_defect(image, camera)

    # Check that the specific pixels have been set to the defect value
    for pixel in specific_pixels:
        assert image[pixel[0], pixel[1]] == 128


def test_pixel_defect_overwrite_pixel_values(camera, image):
    defect = ConstantPixelDefect(name="test_defect", rate=0.1, seed=42)
    defect.set_pixels(defect._select_random_pixels(camera), camera)

    # Create a copy of the image to modify
    modified_image = image.copy()
    defect._overwrite_pixel_values(modified_image, defect.pixels, 255)

    # Check that the specified pixels have been modified
    for pixel in defect.pixels:
        assert modified_image[pixel[0], pixel[1]] == 255


def test_pixel_defect_check_pixel_bounds_valid(camera):
    defect = ConstantPixelDefect(name="test_defect", rate=0.1, seed=42)
    valid_pixels = np.array([[0, 0], [camera.height - 1, camera.width - 1]])
    defect._check_pixel_bounds(
        valid_pixels, camera.height, camera.width, defect.name
    )  # Should not raise


def test_pixel_defect_check_pixel_bounds_invalid(camera):
    defect = ConstantPixelDefect(name="test_defect", rate=0.1, seed=42)
    invalid_pixels = np.array([[camera.height, camera.width]])  # Out of bounds
    with pytest.raises(ValueError):
        defect._check_pixel_bounds(
            invalid_pixels, camera.height, camera.width, defect.name
        )


def test_constant_pixel_defect_introduce(camera, image):
    defect = ConstantPixelDefect(name="constant_defect", rate=0.1, value=1024)
    defect.set_pixels(defect._select_random_pixels(camera), camera)
    modified_image = defect.introduce_pixel_defect(image.copy(), camera)

    # Check that the defect pixels have the correct value
    for pixel in defect.pixels:
        assert modified_image[pixel[0], pixel[1]] == 1024


def test_constant_pixel_defect_no_defect_pixels(camera, image):
    defect = ConstantPixelDefect(name="constant_defect", rate=0.0, value=255)
    defect.set_pixels(defect._select_random_pixels(camera), camera)
    modified_image = defect.introduce_pixel_defect(image.copy(), camera)

    # Check that the image remains unchanged
    assert np.array_equal(modified_image, image)


def test_constant_pixel_defect_invalid_pixel_selection(camera):
    defect = ConstantPixelDefect(name="constant_defect", rate=0.1, value=255)
    with pytest.raises(ValueError):
        defect.set_pixels(np.array([[camera.height, camera.width]]), camera)


def test_telegraphic_pixel_defect_initialization():
    defect = TelegraphicPixelDefect(name="test_defect", rate=0.1, value=1024, dim=0)
    assert defect.name == "test_defect"
    assert defect.rate == 0.1
    assert defect.value == 1024
    assert defect.dim == 0


def test_telegraphic_pixel_defect_introduce(camera, image):
    defect = TelegraphicPixelDefect(name="test_defect", rate=0.1, value=1024, dim=0)
    # defect.set_pixels(defect._select_random_pixel(camera), camera)
    modified_image = defect.introduce_pixel_defect(image.copy(), camera)

    # Check that the defect pixels have the correct value
    for pixel in defect.pixels:
        assert modified_image[pixel[0], pixel[1]] == 1024


def test_telegraphic_pixel_defect_no_defect_pixels(camera, image):
    defect = TelegraphicPixelDefect(name="test_defect", rate=0.0, value=255, dim=0)
    modified_image = defect.introduce_pixel_defect(image.copy(), camera)

    assert np.array_equal(modified_image, image)


def test_telegraphic_pixel_defect_zero_defect_pixels(camera, image):
    defect = TelegraphicPixelDefect(name="test_defect", rate=0.0, value=255, dim=0)
    modified_image = defect.introduce_pixel_defect(image.copy(), camera)

    assert np.array_equal(modified_image, image)


def test_random_noise_pixel_defect_initialization():
    defect = RandomNoisePixelDefect(
        name="test_noise", rate=0.1, noise_level=10.0, distribution="normal"
    )
    assert defect.name == "test_noise"
    assert defect.rate == 0.1
    assert defect.noise_level == 10.0
    assert defect.distribution == "normal"


def test_random_noise_pixel_defect_introduce(camera, image):
    defect = RandomNoisePixelDefect(
        name="test_noise", rate=0.1, noise_level=10.0, distribution="normal"
    )
    modified_image = defect.introduce_pixel_defect(image.copy(), camera)

    assert np.sum(np.abs(modified_image - image)) > 0


def test_random_noise_pixel_defect_distribution(camera, image):
    defect = RandomNoisePixelDefect(
        name="test_noise", rate=0.1, noise_level=10.0, distribution="normal"
    )
    defect.set_pixels(defect._select_random_pixels(camera), camera)
    modified_image = defect.introduce_pixel_defect(image.copy(), camera)

    assert np.sum(np.abs(modified_image - image)) > 0


def test_random_noise_pixel_defect_clipping(camera, image):
    defect = RandomNoisePixelDefect(
        name="test_noise", rate=0.1, noise_level=1000.0, distribution="normal"
    )
    defect.set_pixels(defect._select_random_pixels(camera), camera)
    modified_image = defect.introduce_pixel_defect(image.copy(), camera)

    assert np.all(modified_image <= camera.max_adu)



import pytest

from cabaret import Camera, generate_image


def test_simple():
    camera = Camera(width=100, height=100)
    ra, dec = 323.36152, -0.82325
    _ = generate_image(ra, dec, 1, camera=camera)


def test_timeout():
    camera = Camera(width=100, height=100)
    ra, dec = 323.36152, -0.82325

    with pytest.raises(TimeoutError):
        _ = generate_image(ra, dec, 1, camera=camera, timeout=0.1)



from datetime import UTC, datetime

import pytest

from cabaret.camera import Camera
from cabaret.focuser import Focuser
from cabaret.observatory import Observatory
from cabaret.site import Site
from cabaret.sources import Sources
from cabaret.telescope import Telescope


def test_observatory_initialization():
    observatory = Observatory()
    assert observatory.name == "Observatory"
    assert isinstance(observatory.camera, Camera)
    assert isinstance(observatory.focuser, Focuser)
    assert isinstance(observatory.telescope, Telescope)
    assert isinstance(observatory.site, Site)


def test_observatory_post_init():
    observatory = Observatory(
        camera={"name": "test_camera"},
        telescope={},
        site={},
    )
    assert isinstance(observatory.camera, Camera)
    assert isinstance(observatory.telescope, Telescope)
    assert isinstance(observatory.site, Site)


def test_generate_image_from_sources():
    observatory = Observatory()

    sources = Sources.from_arrays(
        ra=[10.64, 10.68], dec=[10.68, 41.22], fluxes=[169435.6, 52203.9]
    )
    img = observatory.generate_image(
        ra=sources.ra.deg.mean(),
        dec=sources.dec.deg.mean(),
        exp_time=10,
        seed=0,
        sources=sources,
    )
    assert img is not None


def test_generate_image():
    observatory = Observatory()
    dateobs = datetime.now(UTC)
    img = observatory.generate_image(
        ra=12.3323, dec=30.4343, exp_time=10, dateobs=dateobs, seed=0
    )
    assert img is not None


def test_to_dict():
    observatory = Observatory()
    obs_dict = observatory.to_dict()
    assert isinstance(obs_dict, dict)
    assert obs_dict["name"] == "Observatory"


def test_from_dict():
    config = {
        "camera": {"name": "test_camera"},
        "telescope": {},
        "site": {},
    }
    observatory = Observatory.from_dict(config)
    assert isinstance(observatory, Observatory)
    assert observatory.camera.name == "test_camera"


def test_save_to_yaml(tmp_path):
    observatory = Observatory()
    file_path = tmp_path / "observatory.yaml"
    observatory.save_to_yaml(file_path)
    assert file_path.exists()


def test_load_from_yaml(tmp_path):
    file_path = tmp_path / "observatory.yaml"
    Observatory(name="test", camera={"name": "test_camera"}).save_to_yaml(file_path)
    observatory = Observatory.load_from_yaml(file_path)
    assert isinstance(observatory, Observatory)
    assert observatory.name == "test"
    assert observatory.camera.name == "test_camera"


if __name__ == "__main__":
    pytest.main()



from concurrent.futures import ThreadPoolExecutor, TimeoutError
from datetime import datetime

import numpy as np
from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.table import Table
from astropy.units import Quantity
from astroquery.gaia import Gaia

from cabaret.sources import Sources


def tmass_mag_to_photons(mags: np.ndarray) -> np.ndarray:
    """Convert 2MASS J magnitudes to photon fluxes at mag 0.

    Reference: https://lweb.cfa.harvard.edu/~dfabricant/huchra/ay145/mags.html
    Returns photons/sec/m^2 for each magnitude.
    """
    Jy = 1.51e7  # [photons sec^-1 m^-2 (dlambda/lambda)^-1]
    photons = 0.16 * 1600 * Jy  # [photons sec^-1 m^-2] at mag 0
    return photons * 10 ** (-0.4 * mags)


def gaia_launch_job_with_timeout(query, timeout=None, **kwargs) -> Table:
    """
    Launch a Gaia job and return its results, optionally enforcing a timeout.

    Parameters
    ----------
    query : str
        The query string passed to Gaia.launch_job.
    timeout : float or None, optional
        Maximum number of seconds to wait for Gaia.launch_job to complete.
        If None, the job is run on the main thread (no thread overhead).
    **kwargs
        Additional keyword arguments forwarded to Gaia.launch_job.

    Returns
    -------
    object
        The result returned by job.get_results().

    Raises
    ------
    TimeoutError
        If `timeout` is not None and the call does not complete within `timeout`.
    """
    # Run directly on the main thread when no timeout is requested to avoid
    # unnecessary thread creation and to preserve original callstacks/tracebacks.
    if timeout is None:
        job = Gaia.launch_job(query, **kwargs)
        return job.get_results()

    with ThreadPoolExecutor() as executor:
        future = executor.submit(Gaia.launch_job, query, **kwargs)
        try:
            job = future.result(timeout=timeout)
            return job.get_results()
        except TimeoutError:
            raise TimeoutError("Gaia query timed out.")


def gaia_query(
    center: tuple[float, float] | SkyCoord,
    fov: float | Quantity,
    limit: int = 100000,
    circular: bool = True,
    tmass: bool = False,
    timeout: float | None = None,
) -> Table:
    """Query Gaia and return the raw Astropy Table.

    Example
    -------
    >>> from cabaret.queries import gaia_table_query
    >>> from astropy.coordinates import SkyCoord
    >>> center = SkyCoord(ra=10.68458, dec=41.26917, unit='deg')
    >>> table = gaia_table_query(center, fov=0.1, limit=10, timeout=30)
    """
    if isinstance(center, SkyCoord):
        ra = center.ra.deg
        dec = center.dec.deg
    else:
        ra, dec = center

    if not isinstance(fov, u.Quantity):
        fov = u.Quantity(fov, u.deg)

    if fov.ndim == 1:
        ra_fov, dec_fov = fov.to(u.deg).value
    else:
        ra_fov = dec_fov = fov.to(u.deg).value

    radius = np.max([ra_fov, dec_fov]) / 2

    select_cols = [
        "gaia.ra",
        "gaia.dec",
        "gaia.pmra",
        "gaia.pmdec",
        "gaia.phot_rp_mean_flux",
    ]
    joins = []
    where = []
    order_by = "gaia.phot_rp_mean_flux DESC"

    if tmass:
        select_cols.append("tmass.j_m")
        joins.extend(
            [
                "INNER JOIN gaiadr2.tmass_best_neighbour AS tmass_match "
                + "ON tmass_match.source_id = gaia.source_id",
                "INNER JOIN gaiadr1.tmass_original_valid AS tmass "
                + "ON tmass.tmass_oid = tmass_match.tmass_oid",
            ]
        )
        order_by = "tmass.j_m"

    if circular:
        where.append(
            f"1=CONTAINS(POINT('ICRS', {ra}, {dec}), "
            f"CIRCLE('ICRS', gaia.ra, gaia.dec, {radius}))"
        )
    else:
        where.append(
            f"gaia.ra BETWEEN {ra - ra_fov / 2} AND {ra + ra_fov / 2} "
            f"AND gaia.dec BETWEEN {dec - dec_fov / 2} AND {dec + dec_fov / 2}"
        )

    select_clause = ", ".join(select_cols)
    joins_clause = "\n".join(joins)
    where_clause = " AND ".join(where)

    query = f"""
    SELECT TOP {limit} {select_clause}
    FROM gaiadr2.gaia_source AS gaia
    {joins_clause}
    WHERE {where_clause}
    ORDER BY {order_by}
    """

    table = gaia_launch_job_with_timeout(query, timeout=timeout)
    return table


def apply_proper_motion(table: Table, dateobs: datetime):
    """
    Apply proper motion correction to RA and DEC columns for the given observation date.
    """
    dateobs_frac = dateobs.year + (dateobs.timetuple().tm_yday - 1) / 365.25  # type: ignore
    years = dateobs_frac - 2015.5  # type: ignore
    table["ra"] += years * table["pmra"] / 1000 / 3600
    table["dec"] += years * table["pmdec"] / 1000 / 3600
    return table


def get_gaia_sources(
    center: tuple[float, float] | SkyCoord,
    fov: float | Quantity,
    limit: int = 100000,
    circular: bool = True,
    tmass: bool = False,
    dateobs: datetime | None = None,
    timeout: float | None = None,
) -> Sources:
    """
    Query the Gaia archive to retrieve the RA-DEC coordinates of stars
    within a given field-of-view (FOV) centered on a given sky position.

    Parameters
    ----------
    center : tuple or astropy.coordinates.SkyCoord
        The sky coordinates of the center of the FOV.
        If a tuple is given, it should contain the RA and DEC in degrees.
    fov : float or astropy.units.Quantity
        The field-of-view of the FOV in degrees. If a float is given,
        it is assumed to be in degrees.
    limit : int, optional
        The maximum number of sources to retrieve from the Gaia archive.
        By default, it is set to 10000.
    circular : bool, optional
        Whether to perform a circular or a rectangular query.
        By default, it is set to True.
    tmass : bool, optional
        Whether to retrieve the 2MASS J magnitudes catelog.
        By default, it is set to False.
    dateobs : datetime.datetime, optional
        The date of the observation. If given, the proper motions of the sources
        will be taken into account. By default, it is set to None.
    timeout : float, optional
        The maximum time to wait for the Gaia query to complete, in seconds.
        If None, there is no timeout. By default, it is set to None.

    Returns
    -------
    np.ndarray
        A tuple containing:
        - An array of shape (n, 2) containing the RA-DEC coordinates
          of the retrieved sources in degrees.
        - An array of shape (n,) containing the fluxes of the retrieved sources.

    Notes
    -----
    If `tmass` is True, the fluxes are calculated from the 2MASS J-band magnitudes, see
    `cabaret.queries.tmass_mag_to_photons`.

    Raises
    ------
    ImportError
        If the astroquery package is not installed.

    Examples
    --------
    >>> from cabaret.queries import get_gaia_sources
    >>> from astropy.coordinates import SkyCoord
    >>> center = SkyCoord(ra=10.68458, dec=41.26917, unit='deg')
    >>> sources = get_gaia_sources(center, fov=0.1, timeout=30)
    """
    table = gaia_query(
        center=center,
        fov=fov,
        limit=limit,
        circular=circular,
        tmass=tmass,
        timeout=timeout,
    )

    if dateobs is not None:
        table = apply_proper_motion(table, dateobs)

    if tmass:
        fluxes = tmass_mag_to_photons(table["j_m"].value.data)
    else:
        fluxes = table["phot_rp_mean_flux"].value.data
    table.remove_rows(np.isnan(fluxes))
    fluxes = fluxes[~np.isnan(fluxes)]

    return Sources.from_arrays(
        ra=table["ra"].value.data,
        dec=table["dec"].value.data,
        fluxes=fluxes,
    )



from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Literal

import numpy as np
import numpy.random
from astropy.wcs import WCS


@dataclass
class Camera:
    name: str = "gaia-camera-simulated"
    width: int = 1024  # pixels
    height: int = 1024  # pixels
    bin_x: int = 1  # binning factor in x
    bin_y: int = 1  # binning factor in y
    pitch: float = 13.5  # pixel pitch, microns
    plate_scale: float | None = (
        None  # arcsec/pixel (calculated from pitch+telescope if None)
    )
    max_adu: int = 2**16 - 1  # maximum ADU value
    well_depth: int = 2**16 - 1  # electrons
    bias: int = 300  # ADU
    gain: float = 1.0  # e-/ADU
    read_noise: float = 6.2  # e-
    dark_current: float = 0.2  # e-/s
    average_quantum_efficiency: float = 0.8  # fraction
    pixel_defects: dict = field(default_factory=dict)

    def __post_init__(self):
        if self.pixel_defects:
            self.pixel_defects = {
                key: (
                    self._create_pixel_defect(key, **value)
                    if isinstance(value, dict)
                    else value
                )
                for key, value in self.pixel_defects.items()
            }

    def _create_pixel_defect(
        self,
        name: str,
        type: Literal["constant", "telegraphic", "noise"] = "constant",
        **kwargs,
    ) -> "PixelDefect":
        defect_classes = {
            "constant": ConstantPixelDefect,
            "telegraphic": TelegraphicPixelDefect,
            "noise": RandomNoisePixelDefect,
        }
        defect_type = type

        if defect_type not in defect_classes:
            raise ValueError(f"Unknown pixel defect type for {name}")

        return defect_classes[defect_type](name=name, **kwargs)

    @classmethod
    def create_ideal_camera(cls, **kwargs) -> "Camera":
        parameters = {
            "read_noise": 0,
            "dark_current": 0,
            "average_quantum_efficiency": 1.0,
            "bias": 0,
            "gain": 1.0,
        }
        return cls(**(parameters | kwargs))

    def get_wcs(self, center):
        if self.plate_scale is None:
            raise ValueError("plate_scale must be set to compute WCS.")

        wcs = WCS(naxis=2)
        wcs.wcs.cdelt = [-self.plate_scale / 3600, -self.plate_scale / 3600]
        wcs.wcs.cunit = ["deg", "deg"]
        wcs.wcs.crpix = [int(self.width / 2), int(self.height / 2)]
        wcs.wcs.crval = [center.ra.deg, center.dec.deg]
        wcs.wcs.ctype = ["RA---TAN", "DEC--TAN"]
        return wcs

    def _create_blank_image(self) -> np.ndarray:
        """Create a blank image filled with zeros for testing."""
        return np.zeros((self.height, self.width), dtype=np.uint16)


@dataclass
class PixelDefect(ABC):
    name: str = "defect"
    rate: float = 0.0
    seed: int = 0
    _rng: numpy.random.Generator | None = None
    _pixels: np.ndarray | None = None

    @property
    def pixels(self) -> np.ndarray:
        if self._pixels is None:
            raise ValueError(f"{self.name} pixels are not defined.")
        return self._pixels

    @property
    def rng(self) -> numpy.random.Generator:
        if self._rng is None:
            self._rng = numpy.random.default_rng(self.seed)

        return self._rng

    @abstractmethod
    def introduce_pixel_defect(self, image: np.ndarray, camera: Camera):
        """Introduce the defect into the image.

        Parameters
        ----------
        image : np.ndarray
            The image to which the defect will be introduced.
        camera : Camera
            The camera to which the defect applies.
        """
        raise NotImplementedError

    def set_pixels(self, pixels: np.ndarray, camera: Camera):
        """Set the pixels for the defect.

        Parameters
        ----------
        pixels : np.ndarray
            The pixel coordinates of the defect.
        camera : Camera
            The camera to which the defect applies.
        """
        self._check_pixel_bounds(pixels, camera.height, camera.width, self.name)
        self._pixels = pixels

    def number_of_defect_pixels(self, camera: Camera) -> int:
        return int(round(self.rate * camera.width * camera.height))

    def _select_random_pixels(self, camera: Camera) -> np.ndarray:
        return self.rng.integers(
            [camera.height, camera.width],
            size=(self.number_of_defect_pixels(camera), 2),
        )

    @staticmethod
    def _overwrite_pixel_values(
        image: np.ndarray, pixels: np.ndarray, pixel_values: int | np.ndarray
    ):
        if isinstance(pixel_values, np.ndarray):
            if not pixel_values.size == pixels.shape[0]:
                raise ValueError("Pixel values must match the number of pixels.")

        image[pixels[:, 0], pixels[:, 1]] = pixel_values

    @staticmethod
    def _check_pixel_bounds(pixels, height: int, width: int, name: str):
        if pixels is None:
            raise ValueError(f"{name} pixels are not defined.")
        if isinstance(pixels, list):
            pixels = np.array(pixels)
        if not pixels.ndim == 2 or not pixels.shape[1] == 2:
            raise ValueError(
                f"{name} pixels must be a numpy array of (x, y) tuples."
                f"Got shape {pixels.shape} instead."
            )
        if np.any(pixels[:, 0] >= height) or np.any(pixels[:, 1] >= width):
            raise ValueError(f"{name} pixels are outside the frame.")


@dataclass
class ConstantPixelDefect(PixelDefect):
    value: int = 0
    seed: int = 0

    def introduce_pixel_defect(self, image, camera):
        if self._pixels is None:
            self.set_pixels(self._select_random_pixels(camera), camera)

        self._overwrite_pixel_values(image, self.pixels, self.value)

        return image


@dataclass
class TelegraphicPixelDefect(PixelDefect):
    value: int = 0
    seed: int = 0
    dim: int = 0
    _lines: np.ndarray | None = None

    def introduce_pixel_defect(self, image, camera):
        if self._lines is None:
            self.set_lines(self._select_random_lines(camera, self.dim), camera)

        self._overwrite_pixel_values(image, self.pixels, self.value)

        return image

    @property
    def lines(self) -> np.ndarray:
        if self._lines is None:
            raise ValueError(f"{self.name} lines are not defined.")
        return self._lines

    def set_lines(self, lines: np.ndarray | list, camera: Camera):
        self._lines = np.array(lines)

        # Set pixels based on the selected lines
        if self.dim == 0:
            if not np.all(self._lines < camera.height):
                raise ValueError("Selected lines are outside the frame.")
            X, Y = np.meshgrid(np.arange(camera.width), self._lines)
        else:
            if not np.all(self._lines < camera.width):
                raise ValueError("Selected lines are outside the frame.")
            X, Y = np.meshgrid(self._lines, np.arange(camera.width))

        self.set_pixels(np.column_stack((X.ravel(), Y.ravel())), camera)

    def _select_random_lines(self, camera: Camera, dim: int = 0) -> np.ndarray:
        line_length = camera.width if dim == 0 else camera.height
        number_of_lines = self.number_of_defect_pixels(camera) // line_length
        selected_lines = self.rng.integers(line_length, size=(number_of_lines))
        return selected_lines


@dataclass
class RandomNoisePixelDefect(PixelDefect):
    noise_level: float = 10.0  # Standard deviation for Gaussian noise
    distribution: Literal["normal", "poisson"] = "normal"

    def introduce_pixel_defect(self, image, camera, seed: int | None = None):
        if seed is not None:
            self._rng = numpy.random.default_rng(seed)

        if self._pixels is None:
            self.set_pixels(self._select_random_pixels(camera), camera)

        if self.distribution == "poisson":
            noise = self.noise_level * self.rng.poisson(size=self.pixels.shape[0])
        elif self.distribution == "normal":
            noise = self.noise_level * self.rng.normal(size=self.pixels.shape[0])
        else:
            raise ValueError("Unknown noise distribution.")

        image[self.pixels[:, 0], self.pixels[:, 1]] += noise
        image = np.clip(image, 0, camera.max_adu)

        return image



from dataclasses import asdict, dataclass, field
from datetime import UTC, datetime
from pathlib import Path

import numpy.random

from cabaret.camera import Camera
from cabaret.focuser import Focuser
from cabaret.image import generate_image
from cabaret.site import Site
from cabaret.sources import Sources
from cabaret.telescope import Telescope


@dataclass
class Observatory:
    """Observatory configuration.

    Parameters
    ----------
    name : str, optional
        Observatory name.
    camera : Camera, dict, optional
        Camera configuration.
    focuser: Focuser, dict, optional
        Focuser configuration.
    telescope : Telescope, dict, optional
        Telescope configuration.
    site : Site, dict, optional
        Site configuration.

    Examples
    --------
    >>> from datetime import datetime, UTC
    >>> from cabaret.observatory import Observatory
    >>> observatory = Observatory()
    >>> dateobs = datetime.now(UTC)

    Query Gaia for sources and generate an image:
    >>> image = observatory.generate_image(
    ...     ra=12.3323, dec=30.4343, exp_time=10, dateobs=dateobs, seed=0
    ... )

    Or using a set of predefined sources:
    >>> from cabaret.sources import Sources
    >>> sources = Sources.from_arrays(
    ...     ra=[10.64, 10.68], dec=[10.68, 41.22], fluxes=[169435.6, 52203.9]
    ... )
    >>> img = observatory.generate_image(
    ...     ra=sources.ra.deg.mean(),
    ...     dec=sources.dec.deg.mean(),
    ...     exp_time=10,
    ...     seed=0,
    ...     sources=sources,
    ... )
    """

    name: str = "Observatory"
    camera: Camera = field(default_factory=Camera)
    focuser: Focuser = field(default_factory=Focuser)
    telescope: Telescope = field(default_factory=Telescope)
    site: Site = field(default_factory=Site)

    def __post_init__(self):
        if isinstance(self.camera, dict):
            self.camera = Camera(**self.camera)
        if isinstance(self.focuser, dict):
            self.focuser = Focuser(**self.focuser)
        if isinstance(self.telescope, dict):
            self.telescope = Telescope(**self.telescope)
        if isinstance(self.site, dict):
            self.site = Site(**self.site)

        if not isinstance(self.camera, Camera):
            raise ValueError("camera must be an instance of Camera.")
        if not isinstance(self.focuser, Focuser):
            raise ValueError("focuser must be an instance of Focuser.")
        if not isinstance(self.telescope, Telescope):
            raise ValueError("telescope must be an instance of Telescope.")
        if not isinstance(self.site, Site):
            raise ValueError("site must be an instance of Site.")

    def generate_image(
        self,
        ra: float,
        dec: float,
        exp_time: float,
        dateobs: datetime = datetime.now(UTC),
        light: int = 1,
        tmass: bool = True,
        n_star_limit: int = 2000,
        rng: numpy.random.Generator = numpy.random.default_rng(),
        seed: int | None = None,
        timeout: float | None = None,
        sources: Sources | None = None,
    ) -> numpy.ndarray:
        """Generate a simulated image of the sky.

        Parameters
        ----------
        ra : float
            Right ascension of the center of the image in degrees.
        dec : float
            Declination of the center of the image in degrees.
        exp_time : float
            Exposure time in seconds.
        dateobs : datetime, optional
            Observation date and time in UTC.
        light : int, optional
            Light pollution level (1-5).
        tmass : bool, optional
            Include 2MASS stars in the image.
        n_star_limit : int, optional
            Maximum number of stars to include in the image.
        rng : numpy.random.Generator, optional
            Random number generator.
        seed : int, optional
            Random number generator seed.
        timeout : float, optional
            The maximum time to wait for the Gaia query to complete, in seconds.
            If None, there is no timeout. By default, it is set to None.
        sources : Sources, optional
            A collection of sources with their sky coordinates and fluxes.
            If provided, these sources will be used instead of querying Gaia.
        """
        return generate_image(
            ra=ra,
            dec=dec,
            exp_time=exp_time,
            dateobs=dateobs,
            light=light,
            camera=self.camera,
            focuser=self.focuser,
            telescope=self.telescope,
            site=self.site,
            tmass=tmass,
            n_star_limit=n_star_limit,
            rng=rng,
            seed=seed,
            timeout=timeout,
            sources=sources,
        )

    def to_dict(self) -> dict:
        return asdict(self)

    @classmethod
    def from_dict(cls, config) -> "Observatory":
        return cls(
            name=config.get("name", "Observatory"),
            camera=Camera(**config["camera"]),
            focuser=Focuser(**config.get("focuser", {})),
            telescope=Telescope(**config["telescope"]),
            site=Site(**config["site"]),
        )

    @classmethod
    def load_from_yaml(cls, file_path: str | Path) -> "Observatory":
        """Load Observatory configuration from a YAML file."""
        try:
            import yaml

            with open(file_path) as f:
                config = yaml.safe_load(f)

            return cls.from_dict(config)

        except ImportError:
            raise ImportError(
                "Please install PyYAML to load Observatory configuration from YAML."
            )
        except FileNotFoundError:
            raise FileNotFoundError(f"File not found: {file_path}")
        except Exception as e:
            raise Exception(f"Error loading Observatory configuration: {e}")

    def save_to_yaml(self, file_path: str | Path):
        try:
            import yaml

            with open(file_path, "w") as f:
                yaml.dump(self.to_dict(), f)

        except ImportError:
            raise ImportError(
                "Please install PyYAML to save Observatory configuration to YAML."
            )
        except Exception as e:
            raise Exception(f"Error saving Observatory configuration: {e}")



import logging
from datetime import UTC, datetime

import numpy as np
import numpy.random
from astropy import units as u
from astropy.coordinates import AltAz, EarthLocation, SkyCoord, get_sun
from astropy.time import Time

from cabaret.camera import Camera
from cabaret.focuser import Focuser
from cabaret.queries import get_gaia_sources
from cabaret.site import Site
from cabaret.sources import Sources
from cabaret.telescope import Telescope

logger = logging.getLogger("cabaret")


def moffat_profile(
    x: np.ndarray,
    y: np.ndarray,
    x0: float,
    y0: float,
    FWHM: float,
    beta: float = 2.5,
) -> np.ndarray:
    # https://nbviewer.org/github/ysbach/AO_2017/blob/master/04_Ground_Based_Concept.ipynb#1.2.-Moffat
    # FWHM =  2 * R * (2**(1/beta) - 1)**0.5

    R = (FWHM / 2) * (1 / (2 ** (1 / beta) - 1) ** 0.5)
    A = (beta - 1) / (np.pi * R**2)

    r_squared = (x - x0) ** 2 + (y - y0) ** 2

    mp = A * (1 + (r_squared / R**2)) ** (-beta)

    mp_sum = np.sum(mp)

    return mp / mp_sum


def generate_star_image_slow(
    pos: np.ndarray,
    fluxes: list,
    FWHM: float,
    frame_size: tuple,
    rng: numpy.random.Generator,
):
    x = np.linspace(0, frame_size[0] - 1, frame_size[0])
    y = np.linspace(0, frame_size[1] - 1, frame_size[1])
    xx, yy = np.meshgrid(x, y)

    image = np.zeros(frame_size).T
    for i, flux in enumerate(fluxes):
        x0 = pos[0][i]
        y0 = pos[1][i]
        star = rng.poisson(flux) * moffat_profile(xx, yy, x0, y0, FWHM)
        image += star

    return image


def generate_star_image(
    pos: np.ndarray,
    fluxes: list,
    FWHM: float,
    frame_size: tuple,
    rng: numpy.random.Generator,
) -> np.ndarray:
    x = np.linspace(0, frame_size[0] - 1, frame_size[0])
    y = np.linspace(0, frame_size[1] - 1, frame_size[1])
    xx, yy = np.meshgrid(x, y)

    render_radius = FWHM * 5  # render 5 FWHM around the star

    image = np.zeros(frame_size).T
    for i, flux in enumerate(fluxes):
        x0 = pos[0][i]
        y0 = pos[1][i]
        if x0 < 0 or x0 >= frame_size[0] or y0 < 0 or y0 >= frame_size[1]:
            # print(f"Star {i} is outside the frame.")
            continue
        x_min, x_max = int(x0 - render_radius), int(x0 + render_radius)
        y_min, y_max = int(y0 - render_radius), int(y0 + render_radius)
        x_min, x_max = max(0, x_min), min(x_max, frame_size[0] - 1)
        y_min, y_max = max(0, y_min), min(y_max, frame_size[1] - 1)

        star = rng.poisson(flux) * moffat_profile(
            xx[y_min : y_max + 1, x_min : x_max + 1],
            yy[y_min : y_max + 1, x_min : x_max + 1],
            x0,
            y0,
            FWHM,
        )
        image[y_min : y_max + 1, x_min : x_max + 1] += star

    return image


def generate_image(
    ra: float,
    dec: float,
    exp_time: float,
    dateobs: datetime = datetime.now(UTC),
    light: int = 1,
    camera: Camera = Camera(),
    focuser: Focuser = Focuser(),
    telescope: Telescope = Telescope(),
    site: Site = Site(),
    tmass: bool = True,
    n_star_limit: int = 2000,
    rng: numpy.random.Generator = numpy.random.default_rng(),
    seed: int | None = None,
    timeout: float | None = None,
    sources: Sources | None = None,
) -> np.ndarray:
    if seed is not None:
        rng = numpy.random.default_rng(seed)

    base = np.ones((camera.height, camera.width)).astype(np.float64)

    base += rng.poisson(base * camera.dark_current * exp_time).astype(np.float64)

    base += rng.normal(0, camera.read_noise, (camera.height, camera.width)).astype(
        np.float64
    )

    if camera.plate_scale is None:
        camera.plate_scale = (
            np.arctan((camera.pitch * 1e-6) / (telescope.focal_length))
            * (180 / np.pi)
            * 3600
        )  # "/pixel

    if telescope.collecting_area is None:
        telescope.collecting_area = np.pi * (telescope.diameter / 2) ** 2  # [m^2]

    if light == 1:
        # call gaia
        center = SkyCoord(ra=ra, dec=dec, unit="deg")

        fovx = (
            (1 / np.abs(np.cos(center.dec.rad)))
            * camera.width
            * camera.plate_scale
            / 3600
        )
        fovy = (
            np.sqrt(2) * camera.height * camera.plate_scale / 3600
        )  # to account for poles, maybe should scale instead

        logger.info("Querying Gaia for sources...")
        # gaias, vals = get_gaia_sources(
        if not isinstance(sources, Sources):
            sources = get_gaia_sources(
                center,
                (fovx * 1.5, fovy * 1.5),
                tmass=tmass,
                dateobs=dateobs,
                limit=n_star_limit,
                timeout=timeout,
            )
        logger.info(f"Found {len(sources)} sources (user set limit of {n_star_limit}).")

        image = base

        if site.latitude is not None and site.longitude is not None:
            logger.info(
                "Since location specified, calculating sunlight brightness"
                " based on sun's position"
            )
            location = EarthLocation(
                lat=site.latitude * u.deg, lon=site.longitude * u.deg
            )
            obs_time = Time(dateobs, scale="utc")

            # Get sun position
            sun = get_sun(obs_time)

            # Transform to altitude/azimuth coordinates for the observatory
            altaz_frame = AltAz(obstime=obs_time, location=location)
            sun_altaz = sun.transform_to(altaz_frame)
            sun_altitude = sun_altaz.alt.degree
            logger.info(f"Sun altitude: {sun_altitude:.5f} deg at {dateobs} UTC")

            a, b, c = (
                np.float64(4533508.655833181),
                np.float64(0.3937301435229289),
                np.float64(-0.7907223506021084),
            )  # calibrated for I+z band in Paranal

            sky_brightness = a * b ** (c * sun_altitude)  # e-/m2/arcsec2/s
            logger.info(f"sky_brightness (e-/m2/arcsec2/s): {sky_brightness}")

            sky_e = (
                sky_brightness * telescope.collecting_area * camera.plate_scale**2
            )  # e-/s
            logger.info(f"sky_e (e-/s): {sky_e}")

            image += rng.poisson(
                np.ones((camera.height, camera.width)).astype(np.float64)
                * sky_e
                * exp_time
            ).astype(np.float64)

        if len(sources) > 0:
            fluxes = (
                sources.fluxes
                * camera.average_quantum_efficiency
                * telescope.collecting_area
                * exp_time
            )  # [electrons]

            # convert gaia stars to pixel coordinates
            wcs = camera.get_wcs(center)
            gaias_pixel = sources.to_pixel(wcs)

            # stars within frame and moffat profile
            stars = generate_star_image(
                gaias_pixel,
                fluxes,
                focuser.seeing_multiplier * site.seeing / camera.plate_scale,
                (camera.width, camera.height),
                rng=rng,
            ).astype(np.float64)  # * flat

            sky_background = (
                site.sky_background * telescope.collecting_area * camera.plate_scale**2
            )  # [e-/s]

            # make base image with sky background
            image = base + rng.poisson(
                np.ones((camera.height, camera.width)).astype(np.float64)
                * sky_background
                * exp_time
            ).astype(np.float64)  # * flat

            image += stars

    else:
        # dark exposure
        image = base

    # convert to adu and add camera's bias
    image = image / camera.gain + camera.bias  # [adu]

    # inject defect pixels
    for defect in camera.pixel_defects.values():
        image = defect.introduce_pixel_defect(image, camera)

    # clip to max adu
    image = np.clip(image, 0, camera.max_adu)

    # make image 16 bit
    image = image.astype(np.uint16)

    return image


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    camera = Camera(width=2000, height=2000)
    telescope = Telescope()
    site = Site(seeing=1.3, sky_background=350)
    exp_time = 0.1  # [s]

    logger.info("Generating image...")

    # example usage
    image = generate_image(
        323.36152,
        -0.82325,
        exp_time=exp_time,
        camera=camera,
        telescope=telescope,
        site=site,
    )

    science = image  # - camera.dark_current / camera.gain * exp_time - camera.bias

    print("Plotting image...")
    med = np.median(science)
    std = np.std(science)
    print(med, std)

    fig, ax = plt.subplots()
    img = ax.imshow(science, cmap="gray", vmin=med - 1 * std, vmax=med + 1 * std)
    cbar = plt.colorbar(img, ax=ax)
    cbar.set_label("ADU")
    plt.show()



from dataclasses import dataclass


@dataclass
class Focuser:
    """A simple focuser model.

    Parameters
    ----------
    position : float
        The current focus position.
    best_position : float
        The best focus position.
    scale : float
        The scale factor for defocus.
    max_seeing_multiplier : float
        The maximum seeing multiplier possible due to defocus.

    Examples
    --------
    >>> from cabaret.focuser import Focuser
    >>> focuser = Focuser()
    >>> focuser.seeing_multiplier(focus_position=10100)
    2.0
    """

    position: float = 10_000
    best_position: float = 10_000
    scale: float = 100
    max_seeing_multiplier: float = 5.0

    @property
    def seeing_multiplier(
        self,
    ) -> float:
        """Factor by which the seeing is increased due to defocus."""
        offset = abs(self.position - self.best_position)
        return min(1 + offset / self.scale, self.max_seeing_multiplier)



from cabaret.focuser import Focuser

BEST_POSITION = 10_000
SCALE = 100


def test_focuser_default_multiplier():
    focuser = Focuser()
    assert focuser.seeing_multiplier == 1.0


def test_focuser_in_focus():
    focuser = Focuser(best_position=BEST_POSITION)
    assert focuser.seeing_multiplier == 1.0


def test_focuser_out_of_focus():
    focuser = Focuser(
        position=BEST_POSITION + SCALE, best_position=BEST_POSITION, scale=SCALE
    )
    # offset = SCALE, multiplier = 1 + SCALE/SCALE = 2.0
    assert focuser.seeing_multiplier == 2.0


def test_focuser_max_multiplier():
    focuser = Focuser(
        position=BEST_POSITION + 5 * SCALE,
        best_position=BEST_POSITION,
        scale=SCALE,
        max_seeing_multiplier=2.0,
    )
    # offset = 5*SCALE, multiplier = 1 + 5*SCALE/SCALE = 6.0, but max is 2.0
    assert focuser.seeing_multiplier == 2.0


def test_focuser_negative_offset():
    focuser = Focuser(
        position=BEST_POSITION - 2 * SCALE, best_position=BEST_POSITION, scale=100
    )
    # offset = 2*SCALE, multiplier = 1 + 2*SCALE/SCALE = 3.0
    assert focuser.seeing_multiplier == 3.0



from dataclasses import dataclass

import numpy as np


@dataclass
class Telescope:
    focal_length: float = 8.0  # meters
    diameter: float = 1.0  # meters
    collecting_area: float | None = None  # square meters

    def __post_init__(self):
        if self.collecting_area is None:
            self.collecting_area = np.pi * (self.diameter / 2) ** 2  # square meters
